import argparse
from collections import OrderedDict
import copy
import glob
from jinja2 import Template
import os
from os.path import isfile, isdir, splitext
from queue import Queue
import re
import shutil
from subprocess import Popen, PIPE, check_output
import textwrap
from time import sleep
import time
from uuid import uuid4


def combine_report(run_path, calibration):
    sphinx_path = "{}/sphinx_rep".format(os.path.abspath(run_path))
    os.makedirs(sphinx_path)
    direntries = os.listdir(run_path)
    
    for entry in direntries:
        
        if isfile("{}/{}".format(run_path, entry)):
            name, ext = splitext("{}".format(entry))
            
            if ext == ".rst":
                comps = name.split("_")                
                if len(comps) >= 2:
                    comps = "_".join(comps[:-1]), comps[-1]
                    group, module = comps
                    if module == "None":
                        module = "All Modules"
                else:
                    group, module = comps[0], "All Modules"
                    
                with open("{}/{}.rst".format(sphinx_path, group), "a") as gfile:
                    title = "{} - {}".format(calibration, module)
                    gfile.write(title + "\n")
                    gfile.write( "=" *len (title) + "\n")
                    gfile.write("\n")
                    with open("{}/{}".format(run_path, entry), "r") as ifile:
                        for line in ifile.readlines():
                            gfile.write(line)
                    gfile.write("\n\n")
        if isdir("{}/{}".format(run_path, entry)):
            shutil.copytree("{}/{}".format(run_path, entry), "{}/{}".format(sphinx_path, entry))
    return sphinx_path

def make_report(run_path, tmp_path, out_path, project, author, version):
    run_path = os.path.abspath(run_path)
    try:
        import subprocess
        subprocess.check_call(["sphinx-quickstart",
                               "--quiet",
                               "--project='{}'".format(project),
                               "--author='{}'".format(author),
                               "-v", str(version),
                               "--suffix=.rst",
                               "--master=index",
                               "--ext-intersphinx",
                               "--ext-mathjax",
                               "--makefile",
                               "--no-batchfile", run_path])

    except subprocess.CalledProcessError:
        raise Exception("Failed to run sphinx-quickbuild. Is sphinx installed?"
                        "Generated simple index.rst instead")
        
    # quickbuild went well we need to edit the index file
    from shutil import move
    
    direntries = os.listdir(run_path)
    files_to_handle = []
    for entry in direntries:
        if isfile("{}/{}".format(run_path, entry)):
            name, ext = splitext("{}".format(entry))            
            if ext == ".rst" and "index" not in name:
                files_to_handle.append(name)

    with open("{}/index.rst.tmp".format(run_path), "w") as mf:
        with open("{}/index.rst".format(run_path), "r") as mfr:
            indexTmp = Template('''
                .. toctree::
                    :maxdepth: 2
                    {% for k in keys %}
                    {{ k }}
                    {%- endfor %}
                ''')
            for line in mfr:
                line = line.replace(".. toctree::", textwrap.dedent(
                        indexTmp.render(keys=files_to_handle)))
                line = line.replace(":maxdepth: 2", "")
                line = line.replace("Documentation", "Calibration")
                mf.write(line)
                cdir = os.getcwd()
                
    os.remove("{}/index.rst".format(run_path))
    move("{}/index.rst.tmp".format(run_path), "{}/index.rst".format(run_path))

    # finally call the make scripts

    os.chdir(run_path)
    try:
        import subprocess
        subprocess.check_call(["make", "latexpdf"])

    except subprocess.CalledProcessError:
        print("Failed to make pdf documentation")
        print("Temp files will not be deleted and "+
              "can be inspected at: {}".format(run_path))
        return
    print("Moving report to final location: {}".format(out_path))
    for f in glob.glob(r'{}/_build/latex/*.pdf'.format(run_path)):
        print(f)
        shutil.copy(f, out_path)
    print("Removing temporary files at: {}".format(tmp_path))
#    shutil.rmtree(tmp_path)
    
        
def finalize(joblist, run_path, out_path, project, calibration, author, version):
    
    
    print("Waiting on jobs to finish: {}".format(joblist))
    while True:
        found_jobs = set()
        output = check_output(['squeue']).decode('utf8')
        for line in output.split("\n"):
            for job in joblist:
                if str(job) in line:
                    found_jobs.add(job)
        if len(found_jobs) == 0:
            break
        sleep(10)
    sphinx_path = combine_report(run_path, calibration)
    make_report(sphinx_path, run_path, out_path, project, author, version)
    
              
        
def parse_runs(runs, return_type=str):
    pruns = []
    if isinstance(runs, str):
        for rcomp in runs_in.split(","):
            if "-" in rcomp:
                start, end = rcomp.split("-")
                pruns += list(range(int(start), int(end)))
            else:
                pruns += [int(rcomp)]
    elif isinstance(runs, (list, tuple)):
        pruns = runs
    else:
        pruns = [runs,]
        
    if return_type is str:
        return ["r{:04d}".format(r) for r in pruns]
    else:
        return pruns
                

def run_prop_seq_from_path(filename):
    run = re.findall(r'.*r([0-9]{4}).*', filename)
    run = run[0] if len(run) else None
    proposal = re.findall(r'.*p([0-9]{6}).*', filename)
    proposal = proposal[0] if len(proposal) else None
    sequence = re.findall(r'.*S([0-9]{5}).*', filename)
    sequence = sequence[0] if len(sequence) else None
    return run, proposal, sequence


def map_modules_from_files(filelist, file_inset, quadrants, modules_per_quad):
    total_sequences = 0
    total_file_size = 0
    module_files = {}
    mod_ids = {}
    for quadrant in range(0, quadrants):
        for module in range(0, modules_per_quad):
            name = "Q{}M{}".format(quadrant + 1, module + 1)
            module_files[name] = Queue()
            num = quadrant * 4 + module
            mod_ids[name] = num
            file_infix = "{}{:02d}".format(file_inset, num)
            for file in filelist:
                if file_infix in file:
                    module_files[name].put(file)
                    total_sequences += 1
                    total_file_size += os.path.getsize(file)
                    
    return module_files, mod_ids, total_sequences, total_file_size


def gain_map_files(in_folder, runs, sequences, file_inset, quadrants, modules_per_quad):
    total_sequences = 0
    total_file_size = 0
    gain_mapped_files = OrderedDict()
    for gain, run in runs.items():
        ginfolder = "{}/{}".format(in_folder, run)
        dirlist = os.listdir(ginfolder)
        file_list = []
        for entry in dirlist:
            #only h5 file
            abs_entry = "{}/{}".format(ginfolder, entry)
            if os.path.isfile(abs_entry) and os.path.splitext(abs_entry)[1] == ".h5":

                if sequences is None:
                    file_list.append(abs_entry)
                else:
                    for seq in sequences:
                        if "{:05d}.h5".format(seq) in abs_entry:
                            file_list.append(os.path.abspath(abs_entry))
                            
        mapped_files, mod_ids, seq, fs = map_modules_from_files(file_list, file_inset,
                                                                quadrants, modules_per_quad)
        total_sequences += seq
        total_file_size += fs
        gain_mapped_files[gain] = mapped_files
    return gain_mapped_files, total_sequences, total_file_size/1e9


# from https://github.com/jupyter/notebook/issues/1000
import json
import os.path
import re
import ipykernel
import requests

try:  # Python 3
    from urllib.parse import urljoin
except ImportError:  # Python 2
    from urlparse import urljoin

try:  # Python 3
    from notebook.notebookapp import list_running_servers
except ImportError:  # Python 2
    import warnings
    from IPython.utils.shimmodule import ShimWarning
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=ShimWarning)
        from IPython.html.notebookapp import list_running_servers


def get_notebook_name():
    """
    Return the full path of the jupyter notebook.
    """
    try:
        kernel_id = re.search('kernel-(.*).json',
                              ipykernel.connect.get_connection_file()).group(1)
        servers = list_running_servers()
        for ss in servers:
            response = requests.get(urljoin(ss['url'], 'api/sessions'),
                                    params={'token': ss.get('token', '')})
            for nn in json.loads(response.text):
                if nn['kernel']['id'] == kernel_id:
                    relative_path = nn['notebook']['path']
                    return relative_path
    except:
        return os.environ.get("CAL_NOTEBOOK_NAME", "Unknown Notebook")
    
    
def get_dir_creation_date(directory, run):
    import os
    import datetime
    creation_time = os.stat("{}/r{:04d}".format(directory, run)).st_ctime 
    creation_time = datetime.datetime.fromtimestamp(creation_time)
    return(creation_time)


all_ready_printed = {}
def get_constant_from_db(device, constant, condition, empty_constant,
                         cal_db_interface, creation_time = None, print_once=True):
    from iCalibrationDB import ConstantMetaData, Constants, Conditions, Detectors, Versions
    
    if device:
        metadata = ConstantMetaData()        
        metadata.calibration_constant = constant
        metadata.calibration_constant_version = Versions.Now(device)
        metadata.detector_condition = condition
        if creation_time is None:
            metadata.calibration_constant_version = Versions.Now(device=device)
        else:
            metadata.calibration_constant_version = Versions.Timespan(device=device,
                                                                      start=creation_time)
        try:
            metadata.retrieve(cal_db_interface)
            if constant.name not in all_ready_printed or not print_once:
                all_ready_printed[constant.name] = True
                print("{} was injected on: {}".format(constant.name, metadata.calibration_constant_version.begin_at))  
            return constant.data
        except:
            return empty_constant
    else:
        return empty_constant

